@@ -1,31 +1,23 @@
-import logging
+import json
 import random
 import re
 import string
 import sys
 import time
+import warnings
 
 import jwt
-import requests
 from dateutil.parser import isoparse
 
 from dohq_artifactory.exception import ArtifactoryException
+from dohq_artifactory.exception import raise_for_status
+from dohq_artifactory.logger import logger
 
 
 def rest_delay():
     time.sleep(0.5)
 
 
-def raise_errors(r):
-    try:
-        r.raise_for_status()
-    except requests.HTTPError as e:
-        if e.response.status_code >= 400:
-            raise ArtifactoryException(e.response.text)
-        else:
-            raise e
-
-
 def _old_function_for_secret(pw_len=16):
     alphabet_lower = "abcdefghijklmnopqrstuvwxyz"
     alphabet_upper = alphabet_lower.upper()
@@ -65,8 +57,14 @@ else:
     generate_password = _new_function_with_secret_module
 
 
+def deprecation(message):
+    warnings.warn(message, DeprecationWarning, stacklevel=2)
+
+
 class AdminObject(object):
+    prefix_uri = "api"
     _uri = None
+    resource_name = "name"
 
     def __init__(self, artifactory):
         self.additional_params = {}
@@ -74,14 +72,15 @@ class AdminObject(object):
         self.name = None
 
         self._artifactory = artifactory.top
+        self.base_url = self._artifactory.drive
         self._auth = self._artifactory.auth
         self._session = self._artifactory.session
 
     def __repr__(self):
-        return "<{self.__class__.__name__} {self.name}>".format(self=self)
+        return f"<{self.__class__.__name__} {getattr(self, self.resource_name)}>"
 
     def __str__(self):
-        return self.name
+        return getattr(self, self.resource_name)
 
     def _create_json(self):
         """
@@ -95,7 +94,9 @@ class AdminObject(object):
         Create object
         :return: None
         """
-        logging.debug("Create {x.__class__.__name__} [{x.name}]".format(x=self))
+        logger.debug(
+            f"Create {self.__class__.__name__} [{getattr(self, self.resource_name)}]"
+        )
         self._create_and_update(self._session.put)
 
     def _create_and_update(self, method):
@@ -105,16 +106,14 @@ class AdminObject(object):
         """
         data_json = self._create_json()
         data_json.update(self.additional_params)
-        request_url = self._artifactory.drive + "/api/{uri}/{x.name}".format(
-            uri=self._uri, x=self
-        )
+        request_url = f"{self.base_url}/{self.prefix_uri}/{self._uri}/{getattr(self, self.resource_name)}"
         r = method(
             request_url,
             json=data_json,
             headers={"Content-Type": "application/json"},
             auth=self._auth,
         )
-        raise_errors(r)
+        raise_for_status(r)
         rest_delay()
         self.read()
 
@@ -133,22 +132,21 @@ class AdminObject(object):
         True if object exist,
         False else
         """
-        logging.debug("Read {x.__class__.__name__} [{x.name}]".format(x=self))
-        request_url = self._artifactory.drive + "/api/{uri}/{x.name}".format(
-            uri=self._uri, x=self
-        )
-        r = self._session.get(
-            request_url,
-            auth=self._auth,
+        logger.debug(
+            f"Read {self.__class__.__name__} [{getattr(self, self.resource_name)}]"
         )
+        request_url = f"{self.base_url}/{self.prefix_uri}/{self._uri}/{getattr(self, self.resource_name)}"
+        r = self._session.get(request_url, auth=self._auth)
         if 404 == r.status_code or 400 == r.status_code:
-            logging.debug(
-                "{x.__class__.__name__} [{x.name}] does not exist".format(x=self)
+            logger.debug(
+                f"{self.__class__.__name__} [{getattr(self, self.resource_name)}] does not exist"
             )
             return False
         else:
-            logging.debug("{x.__class__.__name__} [{x.name}] exist".format(x=self))
-            raise_errors(r)
+            logger.debug(
+                f"{self.__class__.__name__} [{getattr(self, self.resource_name)}] exist"
+            )
+            raise_for_status(r)
             response = r.json()
             self.raw = response
             self._read_response(response)
@@ -160,18 +158,18 @@ class AdminObject(object):
         :return:
         List of objects
         """
-        # logging.debug('List {x.__class__.__name__} [{x.name}]'.format(x=self))
-        request_url = self._artifactory.drive + "/api/{uri}".format(uri=self._uri)
+        logger.debug(f"List {self.__class__.__name__} [{self.name}]")
+        request_url = f"{self.base_url}/{self.prefix_uri}/{self._uri}"
         response = self._session.get(
             request_url,
             auth=self._auth,
         )
         if response.status_code == 200:
-            # logging.debug('{x.__class__.__name__} [{x.name}] does not exist'.format(x=self))
+            logger.debug(f"{self.__class__.__name__} [{self.name}] does not exist")
             json_response = response.json()
-            return [item.get("name") for item in json_response]
+            return [item.get(self.resource_name) for item in json_response]
         else:
-            # logging.debug('{x.__class__.__name__} [{x.name}] exist'.format(x=self))
+            logger.debug(f"{self.__class__.__name__} [{self.name}] exist")
             return "failed"
 
     def update(self):
@@ -179,7 +177,9 @@ class AdminObject(object):
         Update object
         :return: None
         """
-        logging.debug("Create {x.__class__.__name__} [{x.name}]".format(x=self))
+        logger.debug(
+            f"Create {self.__class__.__name__} [{getattr(self, self.resource_name)}]"
+        )
         self._create_and_update(self._session.post)
 
     def delete(self):
@@ -187,15 +187,15 @@ class AdminObject(object):
         Remove object
         :return: None
         """
-        logging.debug("Remove {x.__class__.__name__} [{x.name}]".format(x=self))
-        request_url = self._artifactory.drive + "/api/{uri}/{x.name}".format(
-            uri=self._uri, x=self
+        logger.debug(
+            f"Remove {self.__class__.__name__} [{getattr(self, self.resource_name)}]"
         )
+        request_url = f"{self.base_url}/{self.prefix_uri}/{self._uri}/{getattr(self, self.resource_name)}"
         r = self._session.delete(
             request_url,
             auth=self._auth,
         )
-        raise_errors(r)
+        raise_for_status(r)
         rest_delay()
 
 
@@ -224,7 +224,7 @@ class User(AdminObject):
         self.internal_password_disabled = False
         self._groups = []
 
-        self._lastLoggedIn = None
+        self._last_logged_in = None
         self._realm = None
 
     def _create_json(self):
@@ -255,30 +255,63 @@ class User(AdminObject):
         self.disable_ui_access = response.get("disableUIAccess")
         self.internal_password_disabled = response.get("internalPasswordDisabled")
         self._groups = response.get("groups", [])
-        self._lastLoggedIn = (
+        self._last_logged_in = (
             isoparse(response["lastLoggedIn"]) if response.get("lastLoggedIn") else None
         )
         self._realm = response.get("realm")
 
     @property
     def encryptedPassword(self):
+        """
+        Method for backwards compatibility, see property encrypted_password
+        :return:
+        """
+        deprecation("encryptedPassword is deprecated, use encrypted_password")
+        return self.encrypted_password
+
+    @property
+    def encrypted_password(self):
+        """
+        Get the encrypted password of the authenticated requestor
+        If you authenticate with an API key, the encrypted API key will be returned in the response.
+        :return: (str) encrypted password
+        """
+        encrypted_password = self._authenticated_user_request(
+            api_url="/api/security/encryptedPassword", request_type=self._session.get
+        )
+
+        return encrypted_password
+
+    def _authenticated_user_request(self, api_url, request_type):
+        """
+        Send API request to artifactory to get user security parameters. auth should be provided
+        :param api_url: querying API url
+        :param request_type: session type, GET | POST | PUT | DELETE
+        :return:
+        """
         if self.password is None:
-            raise ArtifactoryException(
-                "Please, set [self.password] before query encrypted password"
-            )
-        logging.debug("User get encrypted password [{x.name}]".format(x=self))
-        request_url = self._artifactory.drive + "/api/security/encryptedPassword"
-        r = self._session.get(
+            raise ArtifactoryException("Please, set [self.password] before querying")
+
+        request_url = self.base_url + api_url
+        r = request_type(
             request_url,
             auth=(self.name, self.password),
         )
-        raise_errors(r)
-        encrypted_password = r.text
-        return encrypted_password
+        raise_for_status(r)
+        return r.text
 
     @property
     def lastLoggedIn(self):
-        return self._lastLoggedIn
+        """
+        Method for backwards compatibility, see property last_logged_in
+        :return:
+        """
+        deprecation("lastLoggedIn is deprecated, use last_logged_in")
+        return self.last_logged_in
+
+    @property
+    def last_logged_in(self):
+        return self._last_logged_in
 
     @property
     def realm(self):
@@ -311,9 +344,95 @@ class User(AdminObject):
     def groups(self):
         self._groups = []
 
+    @property
+    def api_key(self):
+        return self._ApiKeyManager(self)
+
+    class _ApiKeyManager:
+        def __init__(self, user):
+            """
+            :param user: User instance
+            """
+            self._user = user
+            self.url = "/api/security/apiKey"
+
+        def __repr__(self):
+            return self.get() or ""
+
+        def __str__(self):
+            return self.get() or ""
+
+        def get(self):
+            """
+            Get an API key for the current user
+            :return: (str) API key
+            """
+            response = self._user._authenticated_user_request(
+                api_url=self.url,
+                request_type=self._user._session.get,
+            )
+            api_key = json.loads(response).get("apiKey", "")
+
+            return api_key
+
+        def create(self):
+            """
+            Create an API key for the current user.
+            Returns an error if API key already exists - use api_key_regenerate to regenerate API key instead.
+            :return: (str) API key
+            """
+            response = self._user._authenticated_user_request(
+                api_url=self.url,
+                request_type=self._user._session.post,
+            )
+
+            api_key = json.loads(response)["apiKey"]
+
+            return api_key
+
+        def regenerate(self):
+            """
+            Regenerate an API key for the current user
+            :return: (str) API key
+            """
+            if not self.get():
+                raise ArtifactoryException(
+                    "API key does not exist for {}. Please use api_key.create".format(
+                        self._user.name
+                    )
+                )
+
+            response = self._user._authenticated_user_request(
+                api_url=self.url,
+                request_type=self._user._session.put,
+            )
+            api_key = json.loads(response).get("apiKey", "")
+
+            return api_key
+
+        def revoke(self):
+            """
+            Revokes the current user's API key
+            :return: None
+            """
+            self._user._authenticated_user_request(
+                api_url=self.url, request_type=self._user._session.delete
+            )
+
+        def revoke_for_all_users(self):
+            """
+            Revokes all API keys currently defined in the system
+            Requires a privileged user (Admin only)
+            :return: None
+            """
+            self._user._authenticated_user_request(
+                api_url=self.url + "?deleteAll=1",
+                request_type=self._user._session.delete,
+            )
+
 
 class Group(AdminObject):
-    _uri = "groups"
+    _uri = "security/groups"
     _uri_deletion = "security/groups"
 
     def __init__(self, artifactory, name):
@@ -369,10 +488,10 @@ class Group(AdminObject):
         TODO: New entrypoint would go like
         /api/groups/delete and consumes ["list", "of", "groupnames"]
         """
-        logging.debug("Remove {x.__class__.__name__} [{x.name}]".format(x=self))
-        request_url = self._artifactory.drive + "/api/{uri}/{x.name}".format(
-            uri=self._uri_deletion, x=self
+        logger.debug(
+            f"Remove {self.__class__.__name__} [{getattr(self, self.resource_name)}]"
         )
+        request_url = f"{self.base_url}/{self.prefix_uri}/{self._uri_deletion}/{getattr(self, self.resource_name)}"
         r = self._session.delete(request_url, auth=self._auth)
         r.raise_for_status()
         rest_delay()
@@ -382,13 +501,13 @@ class Group(AdminObject):
         Create object
         :return: None
         """
-        logging.debug("Create {x.__class__.__name__} [{x.name}]".format(x=self))
+        logger.debug(
+            f"Create {self.__class__.__name__} [{getattr(self, self.resource_name)}]"
+        )
         data_json = self._create_json()
         data_json.update(self.additional_params)
-        request_url = self._artifactory.drive + "/api/{uri}".format(
-            uri=self._uri, x=self
-        )
-        r = self._session.post(
+        request_url = f"{self.base_url}/{self.prefix_uri}/{self._uri}/{getattr(self, self.resource_name)}"
+        r = self._session.put(
             request_url,
             json=data_json,
             headers={"Content-Type": "application/json"},
@@ -422,7 +541,7 @@ class GenericRepository(AdminObject):
         return self._artifactory.joinpath(self.name)
 
     def _generate_query(self, package):
-        if self.packageType == Repository.DOCKER:
+        if self.package_type == Repository.DOCKER:
             parts = package.split(":")
 
             name = parts[0]
@@ -432,7 +551,7 @@ class GenericRepository(AdminObject):
 
             return {"name": "manifest.json", "path": {"$match": package}}
 
-        if self.packageType == Repository.PYPI and "/" not in package:
+        if self.package_type == Repository.PYPI and "/" not in package:
             operators = {
                 "<=": "$lte",
                 "<": "$lt",
@@ -453,7 +572,7 @@ class GenericRepository(AdminObject):
 
             return {"@pypi.name": {"$match": package}}
 
-        if self.packageType == Repository.MAVEN and "/" not in package:
+        if self.package_type == Repository.MAVEN and "/" not in package:
             package = package.replace("#", ":")
 
             parts = list(package.split(":"))
@@ -473,7 +592,7 @@ class GenericRepository(AdminObject):
             "$or": [
                 {"name": {"$match": package}},
                 {"path": {"$match": package}},
-                {"@{}.name".format(self.packageType): {"$match": package}},
+                {"@{}.name".format(self.package_type): {"$match": package}},
                 {"@build.name": {"$match": package}},
                 {"artifact.module.build.name": {"$match": package}},
             ]
@@ -532,7 +651,7 @@ class GenericRepository(AdminObject):
 
 
 class Repository(GenericRepository):
-    # List packageType from wiki:
+    # List package_type from wiki:
     # https://www.jfrog.com/confluence/display/RTF/Repository+Configuration+JSON#RepositoryConfigurationJSON-application/vnd.org.jfrog.artifactory.repositories.LocalRepositoryConfiguration+json
     ALPINE = "alpine"
     BOWER = "bower"
@@ -550,11 +669,6 @@ class Repository(GenericRepository):
     HELM = "helm"
     IVY = "ivy"
     MAVEN = "maven"
-    SBT = "sbt"
-    HELM = "helm"
-    RPM = "rpm"
-    NUGET = "nuget"
-    GEMS = "gems"
     NPM = "npm"
     NUGET = "nuget"
     PUPPET = "puppet"
@@ -563,21 +677,47 @@ class Repository(GenericRepository):
     SBT = "sbt"
     YUM = "yum"
 
-    # List dockerApiVersion from wiki:
+    # List docker_api_version from wiki:
     V1 = "V1"
     V2 = "V2"
 
     @staticmethod
-    def create_by_type(type: str, artifactory, name):
-        if type == "LOCAL":
+    def create_by_type(repo_type="LOCAL", artifactory=None, name=None, *, type=None):
+        if type is not None:
+            deprecation("'type' argument is deprecated, use 'repo_type'")
+            repo_type = type
+
+        if repo_type == "LOCAL":
             return RepositoryLocal(artifactory, name)
-        elif type == "REMOTE":
+        elif repo_type == "REMOTE":
             return RepositoryRemote(artifactory, name)
-        elif type == "VIRTUAL":
+        elif repo_type == "VIRTUAL":
             return RepositoryVirtual(artifactory, name)
         else:
             return None
 
+    @property
+    def packageType(self):
+        deprecation("packageType is deprecated, use package_type")
+        return self.package_type
+
+    @property
+    def repoLayoutRef(self):
+        deprecation("repoLayoutRef is deprecated, use repo_layout_ref")
+        return self.repo_layout_ref
+
+    @property
+    def dockerApiVersion(self):
+        deprecation("dockerApiVersion is deprecated, use docker_api_version")
+        return self.docker_api_version
+
+    @property
+    def archiveBrowsingEnabled(self):
+        deprecation(
+            "archiveBrowsingEnabled is deprecated, use archive_browsing_enabled"
+        )
+        return self.archive_browsing_enabled
+
 
 class RepositoryLocal(Repository):
     _uri = "repositories"
@@ -590,17 +730,30 @@ class RepositoryLocal(Repository):
         self,
         artifactory,
         name,
-        packageType=Repository.GENERIC,
-        dockerApiVersion=Repository.V1,
-        repoLayoutRef="maven-2-default",
+        package_type=Repository.GENERIC,
+        docker_api_version=Repository.V1,
+        repo_layout_ref="maven-2-default",
+        max_unique_tags=0,
+        *,
+        packageType=None,
+        dockerApiVersion=None,
+        repoLayoutRef=None,
     ):
         super(RepositoryLocal, self).__init__(artifactory)
         self.name = name
         self.description = ""
-        self.packageType = packageType
-        self.repoLayoutRef = repoLayoutRef
-        self.archiveBrowsingEnabled = True
-        self.dockerApiVersion = dockerApiVersion
+        self.package_type = packageType or package_type
+        self.repo_layout_ref = repoLayoutRef or repo_layout_ref
+        self.archive_browsing_enabled = True
+        self.docker_api_version = dockerApiVersion or docker_api_version
+        self.max_unique_tags = max_unique_tags
+
+        if any([packageType, dockerApiVersion, repoLayoutRef]):
+            msg = (
+                "packageType, dockerApiVersion, repoLayoutRef are deprecated, "
+                "use package_type, docker_api_version, repo_layout_ref"
+            )
+            deprecation(msg)
 
     def _create_json(self):
         """
@@ -610,12 +763,12 @@ class RepositoryLocal(Repository):
             "rclass": "local",
             "key": self.name,
             "description": self.description,
-            "packageType": self.packageType,
+            "packageType": self.package_type,
             "notes": "",
             "includesPattern": "**/*",
             "excludesPattern": "",
-            "repoLayoutRef": self.repoLayoutRef,
-            "dockerApiVersion": self.dockerApiVersion,
+            "repoLayoutRef": self.repo_layout_ref,
+            "dockerApiVersion": self.docker_api_version,
             "checksumPolicyType": "client-checksums",
             "handleReleases": True,
             "handleSnapshots": True,
@@ -624,9 +777,15 @@ class RepositoryLocal(Repository):
             "suppressPomConsistencyChecks": True,
             "blackedOut": False,
             "propertySets": [],
-            "archiveBrowsingEnabled": self.archiveBrowsingEnabled,
+            "archiveBrowsingEnabled": self.archive_browsing_enabled,
             "yumRootDepth": 0,
         }
+        """
+        Docker V2 API specific fields
+        """
+        if self.docker_api_version == Repository.V2:
+            data_json["maxUniqueTags"] = self.max_unique_tags
+
         return data_json
 
     def _read_response(self, response):
@@ -643,9 +802,9 @@ class RepositoryLocal(Repository):
 
         self.name = response["key"]
         self.description = response.get("description")
-        self.packageType = response.get("packageType")
-        self.repoLayoutRef = response.get("repoLayoutRef")
-        self.archiveBrowsingEnabled = response.get("archiveBrowsingEnabled")
+        self.package_type = response.get("packageType")
+        self.repo_layout_ref = response.get("repoLayoutRef")
+        self.archive_browsing_enabled = response.get("archiveBrowsingEnabled")
 
 
 class RepositoryVirtual(GenericRepository):
@@ -678,15 +837,26 @@ class RepositoryVirtual(GenericRepository):
         artifactory,
         name,
         repositories=None,
-        packageType=Repository.GENERIC,
+        package_type=Repository.GENERIC,
+        *,
+        packageType=None,
     ):
         super(RepositoryVirtual, self).__init__(artifactory)
         self.name = name
         self.description = ""
         self.notes = ""
-        self.packageType = packageType
+        self.package_type = packageType or package_type
         self.repositories = repositories or []
 
+        if packageType:
+            msg = "packageType is deprecated, use package_type"
+            deprecation(msg)
+
+    @property
+    def packageType(self):
+        deprecation("packageType is deprecated, use package_type")
+        return self.package_type
+
     def _create_json(self):
         """
         JSON Documentation: https://www.jfrog.com/confluence/display/RTF/Repository+Configuration+JSON
@@ -695,7 +865,7 @@ class RepositoryVirtual(GenericRepository):
             "rclass": "virtual",
             "key": self.name,
             "description": self.description,
-            "packageType": self.packageType,
+            "packageType": self.package_type,
             "repositories": self._repositories,
             "notes": self.notes,
         }
@@ -716,7 +886,7 @@ class RepositoryVirtual(GenericRepository):
 
         self.name = response["key"]
         self.description = response.get("description")
-        self.packageType = response.get("packageType")
+        self.package_type = response.get("packageType")
         self._repositories = response.get("repositories")
 
     def add_repository(self, *repos):
@@ -759,19 +929,30 @@ class RepositoryRemote(Repository):
         artifactory,
         name,
         url=None,
-        packageType=Repository.GENERIC,
-        dockerApiVersion=Repository.V1,
-        repoLayoutRef="maven-2-default",
+        package_type=Repository.GENERIC,
+        docker_api_version=Repository.V1,
+        repo_layout_ref="maven-2-default",
+        *,
+        packageType=None,
+        dockerApiVersion=None,
+        repoLayoutRef=None,
     ):
         super(RepositoryRemote, self).__init__(artifactory)
         self.name = name
         self.description = ""
-        self.packageType = packageType
-        self.repoLayoutRef = repoLayoutRef
-        self.archiveBrowsingEnabled = True
-        self.dockerApiVersion = dockerApiVersion
+        self.package_type = packageType or package_type
+        self.repo_layout_ref = repoLayoutRef or repo_layout_ref
+        self.archive_browsing_enabled = True
+        self.docker_api_version = dockerApiVersion or docker_api_version
         self.url = url
 
+        if any([packageType, dockerApiVersion, repoLayoutRef]):
+            msg = (
+                "packageType, dockerApiVersion, repoLayoutRef are deprecated, "
+                "use package_type, docker_api_version, repo_layout_ref"
+            )
+            deprecation(msg)
+
     def _create_json(self):
         """
         JSON Documentation: https://www.jfrog.com/confluence/display/RTF/Repository+Configuration+JSON
@@ -780,12 +961,12 @@ class RepositoryRemote(Repository):
             "rclass": "remote",
             "key": self.name,
             "description": self.description,
-            "packageType": self.packageType,
+            "packageType": self.package_type,
             "notes": "",
             "includesPattern": "**/*",
             "excludesPattern": "",
-            "repoLayoutRef": self.repoLayoutRef,
-            "dockerApiVersion": self.dockerApiVersion,
+            "repoLayoutRef": self.repo_layout_ref,
+            "dockerApiVersion": self.docker_api_version,
             "checksumPolicyType": "client-checksums",
             "handleReleases": True,
             "handleSnapshots": True,
@@ -794,7 +975,7 @@ class RepositoryRemote(Repository):
             "suppressPomConsistencyChecks": True,
             "blackedOut": False,
             "propertySets": [],
-            "archiveBrowsingEnabled": self.archiveBrowsingEnabled,
+            "archiveBrowsingEnabled": self.archive_browsing_enabled,
             "yumRootDepth": 0,
             "url": self.url,
             "debianTrivialLayout": False,
@@ -821,9 +1002,9 @@ class RepositoryRemote(Repository):
 
         self.name = response["key"]
         self.description = response.get("description")
-        self.packageType = response.get("packageType")
-        self.repoLayoutRef = response.get("repoLayoutRef")
-        self.archiveBrowsingEnabled = response.get("archiveBrowsingEnabled")
+        self.package_type = response.get("packageType")
+        self.repo_layout_ref = response.get("repoLayoutRef")
+        self.archive_browsing_enabled = response.get("archiveBrowsingEnabled")
         self.url = response.get("url")
 
 
@@ -1073,7 +1254,7 @@ class Token(AdminObject):
         :return: None
         """
         payload = self._prepare_request()
-        request_url = self._artifactory.drive + "/api/{uri}".format(uri=self._uri)
+        request_url = f"{self.base_url}/{self.prefix_uri}/{self._uri}"
         r = self._session.post(
             request_url,
             data=payload,
@@ -1148,16 +1329,20 @@ class Token(AdminObject):
         True if object exist,
         False else
         """
-        logging.debug("Read {x.__class__.__name__} [{x.name}]".format(x=self))
-        request_url = self._artifactory.drive + "/api/{uri}".format(uri=self._uri)
+        logger.debug(
+            f"Read {self.__class__.__name__} [{getattr(self, self.resource_name)}]"
+        )
+        request_url = f"{self.base_url}/{self.prefix_uri}/{self._uri}"
         r = self._session.get(request_url, auth=self._auth)
         if 404 == r.status_code or 400 == r.status_code:
-            logging.debug(
-                "{x.__class__.__name__} [{x.name}] does not exist".format(x=self)
+            logger.debug(
+                f"{self.__class__.__name__} [{getattr(self, self.resource_name)}] does not exist"
             )
             return False
         else:
-            logging.debug("{x.__class__.__name__} [{x.name}] exist".format(x=self))
+            logger.debug(
+                f"{self.__class__.__name__} [{getattr(self, self.resource_name)}] exist"
+            )
             r.raise_for_status()
             response = r.json()
             self.raw = response
@@ -1173,12 +1358,126 @@ class Token(AdminObject):
         POST security/token/revoke
         revoke (calling it deletion to be consistent with other classes) a token
         """
-        logging.debug("Delete {x.__class__.__name__} [{x.name}]".format(x=self))
-        request_url = self._artifactory.drive + "/api/{uri}".format(
-            uri=self._uri + "/revoke"
+        logger.debug(
+            f"Delete {self.__class__.__name__} [{getattr(self, self.resource_name)}]"
         )
+        request_url = f"{self.base_url}/{self.prefix_uri}/{self._uri}/revoke"
         payload = self._prepare_deletion()
 
         r = self._session.post(request_url, data=payload, auth=self._auth)
         r.raise_for_status()
         rest_delay()
+
+
+class Project(AdminObject):
+    prefix_uri = "access/api"
+    _uri = "v1/projects"
+    resource_name = "project_key"
+
+    def __init__(
+        self,
+        artifactory,
+        project_key,
+        display_name=None,
+        description="",
+        manage_members=True,
+        manage_resources=True,
+        manage_security_assets=True,
+        index_resources=True,
+        allow_ignore_rules=True,
+        storage_quota_bytes=-1,
+        soft_limit=False,
+        storage_quota_email_notification=True,
+    ):
+        self._artifactory = artifactory.top
+        # TODO: What if 'artifactory' is not in 'drive'
+        self.base_url = self._artifactory.drive.rpartition("/artifactory")[0]
+        self._auth = self._artifactory.auth
+        self._session = self._artifactory.session
+
+        self.display_name = display_name
+        self.project_key = project_key
+        self.description = description
+        self.manage_members = manage_members
+        self.manage_resources = manage_resources
+        self.manage_security_assets = manage_security_assets
+        self.index_resources = index_resources
+        self.allow_ignore_rules = allow_ignore_rules
+        self.storage_quota_bytes = storage_quota_bytes
+        self.soft_limit = soft_limit
+        self.storage_quota_email_notification = storage_quota_email_notification
+
+    def create(self):
+        """
+        Create object
+        :return: None
+        """
+        data_json = self._create_json()
+        request_url = self.base_url + "/{prefix_uri}/{uri}".format(
+            prefix_uri=self.prefix_uri, uri=self._uri
+        )
+        r = self._session.post(
+            request_url,
+            json=data_json,
+            headers={"Content-Type": "application/json"},
+            auth=self._auth,
+        )
+        raise_for_status(r)
+        rest_delay()
+        self.read()
+
+    def update(self):
+        """
+        Update object
+        :return: None
+        """
+        data_json = self._create_json()
+        request_url = self.base_url + "/{prefix_uri}/{uri}/{key}".format(
+            prefix_uri=self.prefix_uri,
+            uri=self._uri,
+            key=getattr(self, self.resource_name),
+        )
+        r = self._session.put(
+            request_url,
+            json=data_json,
+            headers={"Content-Type": "application/json"},
+            auth=self._auth,
+        )
+        raise_for_status(r)
+        rest_delay()
+        self.read()
+
+    def _create_json(self):
+        data_json = {
+            "display_name": self.display_name,
+            "project_key": self.project_key,
+            "description": self.description,
+            "admin_privileges": {
+                "manage_members": self.manage_members,
+                "manage_resources": self.manage_resources,
+                "manage_security_assets": self.manage_security_assets,
+                "index_resources": self.index_resources,
+                "allow_ignore_rules": self.allow_ignore_rules,
+            },
+            "storage_quota_bytes": self.storage_quota_bytes,
+        }
+        return data_json
+
+    def _read_response(self, response):
+        self.display_name = response.get("display_name")
+        self.project_key = response.get("project_key")
+        self.description = response.get("description")
+        self.manage_members = response.get("admin_privileges").get("manage_members")
+        self.manage_resources = response.get("admin_privileges").get("manage_resources")
+        self.manage_security_assets = response.get("admin_privileges").get(
+            "manage_security_assets"
+        )
+        self.index_resources = response.get("admin_privileges").get("index_resources")
+        self.allow_ignore_rules = response.get("admin_privileges").get(
+            "allow_ignore_rules"
+        )
+        self.storage_quota_bytes = response.get("storage_quota_bytes")
+        self.soft_limit = response.get("soft_limit")
+        self.storage_quota_email_notification = response.get(
+            "storage_quota_email_notification"
+        )
